/* 
 * A fileselector widget for W toolkit
 * 
 * (w) 1996 by Eero Tamminen
 *
 * $Id: filesel.c,v 1.2 1996/04/04 20:11:04 yak Exp $
 */

#include <stdlib.h>
#include <sys/stat.h>
#include <dirent.h>
#include <string.h>
#include <limits.h>
#include <string.h>
#include <unistd.h>
#include <stdio.h>
#include <Wlib.h>
#include <Wt.h>
#include "toolkit.h"

typedef struct
{
  char dummy;
  void *dummy2;
} pointer_align_test;

/* directory name space allocator configuration defines */
#define POINTER_ALIGN		(sizeof(pointer_align_test)-sizeof(void*))
#define BLOCK_SIZE		1024

/* entry in the read directory */
typedef struct ENTRY
{
  struct ENTRY *next;
  long size;
  int flags;
} Entry;
/* followed by filename */

/* Memory handling / usage:
 *
 * pointer aligning is needed for the structures after strings in the
 * blocks.  One block can contain at least (BLOCK_SIZE / (NAME_MAX +
 * POINTER_ALIGN + sizeof(void*) * 3)) directory entries.  Using
 * sort_entries() will take further sizeof(char*) / entry.
 */


/* stuff in the masked & sorted directory */
typedef struct
{
  char **start;
  int files;
  int items;		/* files + dirs */
  long bytes;
} Directory;


/* directory entry file type flags */
#define FLAG_DIR	1

/* directory / file listbox height */
#define LIST_SHOWN	18

/* lenght of the pattern for the shown filenames on screen */
#define TYPE_WIDTH	7

/* maximum lenghts (at least 30) for certain strings */
#define NAME_WIDTH	30

typedef struct
{
  widget_t w;
  short is_open;
  short queried;
  short is_realized;
  widget_t *shell;
  widget_t *listbox;
  widget_t *type;
  widget_t *info;
  widget_t *name;
  widget_t *path;
  char *memend;				/* last allocated block */
  Entry *start;				/* first directory entry */
  int dirs, files;			/* number of all entries */
  Directory dirinfo;			/* sorted name lists for listboxes */
  char pathbuf[PATH_MAX+1];		/* path parsing buffer */
  void (*action_cb)(widget_t *w, char *path, char *name);
} filesel_widget_t;


/**********************************************************************/

/*
 * glob-style filename matching: does *, ? and []
 */
static int
mask_name (register const char *s, register const char *p)
{
	register int scc;
	int ok, lc;
	int c, cc;

	for (;;) {
		scc = *s++ & 0177;
		switch ((c = *p++)) {
		case '[':
			ok = 0;
			lc = 077777;
			while ((cc = *p++)) {
				if (cc == ']') {
					if (ok)
						break;
					return 0;
				}
				if (cc == '-') {
					if (lc <= scc && scc <= *p++)
						ok++;
				} else
					if (scc == (lc = cc))
						ok++;
			}
			if (cc == 0)
				if (ok)
					p--;
				else
					return 0;
			continue;

		case '*':
			if (!*p)
				return 1;
			s--;
			do {
				if (mask_name(s, p))
					return 1;
			} while (*s++);
			return 0;

		case 0:
			return (scc == 0);

		default:
			if (c != scc)
				return 0;
			continue;

		case '?':
			if (scc == 0)
				return 0;
			continue;
		}
	}
}

static int
locate_first(const char *string, char **list, int items)
{
  int index, len = strlen (string);

  if(!(len && list && items))
    return -1;

  for(index = 0; index < items; index++)
  {
    if (!strncmp (string, *list++, len))
      return index;
  }
  return -1;
}

/**********************************************************************/

/*
 * Directory access --- The POSIX way (I hope <g>)
 */

static void free_entries(filesel_widget_t *w);

/* get directory entry information and copy it into memory */
static inline void get_stats(filesel_widget_t *w, char *name, long *size, int *flags)
{
  struct stat st;
  *flags = 0;

  /* get file information */
  if(stat(name, &st) < 0)
    *size = 0;
  else
  {
    *size = st.st_size;
    if(S_ISDIR(st.st_mode))
      *flags |= FLAG_DIR;
    /* set other flags */
  }
}

/* read all the directory entries + handle memory allocation */
static int read_entries(filesel_widget_t *w, const char *dir)
{
  char *new_block;
  struct dirent *entry;
  Entry *current, *prev;
  int len, length, mem_left, flags;
  long size;
  DIR *dfd;

  if(!(dfd = opendir(dir)))
    return 0;

  if(w->memend)
    free_entries(w);

  if(!(w->memend = malloc(BLOCK_SIZE)))
  {
    closedir(dfd);
    return 0;
  }
  *(char **)w->memend = NULL;			/* this is the first block */
  w->start = (Entry *)(w->memend + sizeof(char *));
  mem_left = BLOCK_SIZE - sizeof(char *);
  current = prev = w->start;
  w->files = w->dirs = 0;

  while((entry = readdir(dfd)))
  {
    /* get directory entry information */
    get_stats(w, entry->d_name, &size, &flags);
    len = strlen(entry->d_name);
    if(flags & FLAG_DIR)
      len++;

    /* takes into account string termination zero */
    length = (sizeof(Entry) + len + POINTER_ALIGN) & ~(POINTER_ALIGN-1);

    /* allocate more memory if needed */
    if(mem_left < length)
    {
      if(!(new_block = malloc(BLOCK_SIZE)))
	break;					/* out of memory */

      *(char **)new_block = w->memend;
      current = (Entry *)(new_block + sizeof(char *));
      mem_left = BLOCK_SIZE - sizeof(char *);
      w->memend = new_block;
    }
    /* set entry information */
    current->size = size;
    current->flags = flags;
    strcpy((char*)(current+1), entry->d_name);
    if(flags & FLAG_DIR)
    {
      new_block = (char*)(current+1) + len;
      *(new_block--) = '\0';
      *new_block = '/';
      w->dirs++;
    }
    else
      w->files++;

    /* set next entry */
    prev->next = current;
    prev = current;
    current = (Entry *)((char *)current + length);
    mem_left -= length;
  }
  prev->next = NULL;
  closedir(dfd);
  return 1;
}

/* compare entry names */
static int name_compare(const void *a, const void *b)
{
  return strcmp(*(char**)a, *(char**)b);
}

/* sort directories and files separately after each other into same list.
 * files are filtered with the 'type' mask.
 * there's no size nor reverse sort as locate relies on list being
 * progressively sorted by name...
 */
static int sort_entries(filesel_widget_t *w, const char *type)
{
  Entry *current = w->start;
  char **dirlist, **filelist;
  long bytes = 0;
  int files;

  if(!(current && type &&  w->dirs + w->files))
    return 0;

  /* not just resort? */
  if(!w->dirinfo.start)
  {
    w->dirinfo.start = malloc((w->dirs + w->files + 1) * sizeof(char*));
    if(!w->dirinfo.start)
      return 0;
  }
  filelist = w->dirinfo.start + w->dirs;
  dirlist = w->dirinfo.start;
  files = 0;

  do
  {
    /* bytecount should probably ignore "." and ".." entries... */
    bytes += current->size;

    if(current->flags & FLAG_DIR)
      *(dirlist++) = (char*)(current+1);
    else
    {
      *filelist = (char*)(current+1);

      /* if no type or type matched, accept the name */
      if(!*type || mask_name(*filelist, type))
      {
        filelist++;
	files++;
      }
    }
    current = current->next;
  } while(current);
  *filelist = NULL;

  /* first sort by name... */
  qsort(w->dirinfo.start,w->dirs, sizeof(char*), name_compare);
  qsort(w->dirinfo.start+w->dirs, files, sizeof(char*), name_compare);
  /* here would be other sorts... */

  w->dirinfo.items = w->dirs + files;
  w->dirinfo.bytes = bytes;
  w->dirinfo.files = files;
  return 1;	/* alles ok */
}

/* free all the allocated spaces */
static void free_entries(filesel_widget_t *w)
{
  char *previous;

  /* free lists */
  if(w->dirinfo.start)
  {
    free(w->dirinfo.start);
    w->dirinfo.start = 0;
  }

  /* free blocks */
  do
  {
    previous = *(char **)w->memend;
    free(w->memend);
    w->memend = previous;
  } while(w->memend);
}


/*
 * Options dialog handling
 */

static void option_cb(widget_t *w, int down)
{
#if 0
  /* fileselector <- shell <- vpane <- hpane <- button */
  filesel_widget_t *w = (filesel_widget_t*)(_w->parent->parent->parent->parent);
  if(down)
    return;
  /* here comes the option dialog... */
#endif
}

/*
 * Fileselector functionality
 */

/* show information for new or re-sorted directory */
static void show_newstuff(filesel_widget_t *w)
{
  char buf[30];

  wt_getstring_class->setopt(w->name, WT_STRING_ADDRESS, "");
  wt_listbox_class->setopt(w->listbox, WT_LIST_ADDRESS, w->dirinfo.start);

  sprintf(buf, "%8ld bytes in %3d files", w->dirinfo.bytes, w->dirinfo.items);
  wt_label_class->setopt(w->info, WT_LABEL, buf);
}

static int get_directory(filesel_widget_t *w, const char *dir)
{
  char *type, *oldpath;
  int len, ok;

  /* change to the directory and read entries */
  oldpath = getcwd(NULL, PATH_MAX);
  chdir(w->pathbuf);
  chdir(dir);			/* stat needs us to be in the directory */

  if((ok = read_entries(w, ".")))
  {
    /* if reading succeeded, show the path */
    getcwd(w->pathbuf, PATH_MAX);
    if((len = strlen(w->pathbuf)) > 1)
    {
      /* not root, add path mark */
      w->pathbuf[len++] = '/';
      w->pathbuf[len] = '\0';
    }
    wt_getstring_class->setopt(w->path, WT_STRING_ADDRESS, w->pathbuf);
  }
  chdir(oldpath);
  free(oldpath);

  if(ok)
  {
    /* mask and sort entries before showing them */
    wt_getstring_class->getopt(w->type, WT_STRING_ADDRESS, &type);
    if(sort_entries(w, type))
    {
      show_newstuff(w);
      return 1;
    }
  }
  wt_listbox_class->setopt(w->listbox, WT_LIST_ADDRESS, "");
  return 0;
}

/* show name and information for a file.
 * *NOTE*: 'name' is supposed to be part of an Entry struct!!!!
 */
static void show_fileinfo(filesel_widget_t *w, char *name, int both)
{
  char buf[16];

  if(both)
    wt_getstring_class->setopt(w->name, WT_STRING_ADDRESS, name);

  sprintf(buf, "%ld bytes", ((Entry *)name - 1)->size);
  wt_label_class->setopt(w->info, WT_LABEL, buf);
}

/* fileselector <- shell <- vpane <- widget */

static void path_cb(widget_t *_w, const char *text, int cursor)
{
  get_directory((filesel_widget_t*)(_w->parent->parent->parent), text);
}

static void listbox_change_cb(widget_t *_w, char *text, int index)
{
  filesel_widget_t *w = (filesel_widget_t*)(_w->parent->parent->parent);
  show_fileinfo(w, text, 1);
}

static void listbox_cb(widget_t *_w, char *text, int index)
{
  filesel_widget_t *w = (filesel_widget_t*)(_w->parent->parent->parent);
  if(index < 0)
    return;
  if(index < w->dirs)
    get_directory(w, text);
  else
    show_fileinfo(w, text, 1);
}

static void route_listbox_key(widget_t *_w, WEVENT *ev)
{
  filesel_widget_t *w = (filesel_widget_t*)(_w->parent->parent->parent);
  /* route keys unknown to listbox into name widget */
  wt_getstring_class->event(w->name, ev);
}

/* check whether we may exit. If yes and exit -> call callback */
static void select_file(filesel_widget_t *w, int exit)
{
  char *name;
  int idx;

  wt_getstring_class->getopt(w->name, WT_STRING_ADDRESS, &name);
  idx = locate_first(name, w->dirinfo.start, w->dirinfo.items);
  if(idx >= 0)
  {
    char *item = w->dirinfo.start[idx];
    if(idx < w->dirs)
    {
      /* change to a directory */
      get_directory(w, item);
      return;
    }
    if(strcmp(name, item))
    {
      /* complete filename and get back */ 
      show_fileinfo(w, item, 1);
      return;
    }
  }
  if(exit && w->action_cb)
    w->action_cb((widget_t*)w, w->pathbuf, name);
}

static void name_change_cb(widget_t *_w, const char *txt, int cursor)
{
  filesel_widget_t *w = (filesel_widget_t*)(_w->parent->parent->parent);
  long idx;

  idx = locate_first(txt, w->dirinfo.start, w->dirinfo.items);
  wt_listbox_class->setopt(w->listbox, WT_CURSOR, &idx);
  if(idx >= 0)
    show_fileinfo(w, w->dirinfo.start[idx], 0);
}

static void name_cb(widget_t *_w, const char *txt, int cursor)
{
  select_file((filesel_widget_t*)(_w->parent->parent->parent), 1);
}

/* fileselector <- shell <- vpane <- hpane <- widget */

static void type_cb(widget_t *_w, const char *mask, int cursor)
{
  filesel_widget_t *w = (filesel_widget_t*)(_w->parent->parent->parent->parent);
  char *type;

  /* resort stuff */
  wt_getstring_class->getopt(w->type, WT_STRING_ADDRESS, &type);
  sort_entries(w, type);
  show_newstuff(w);
}

static void ok_cb(widget_t *_w, int down)
{
  if(!down)
    select_file((filesel_widget_t*)(_w->parent->parent->parent->parent), 1);
}

static void cancel_cb(widget_t *_w, int down)
{
  filesel_widget_t *w = (filesel_widget_t*)(_w->parent->parent->parent->parent);
  if (!down)
    if(w->action_cb)
      w->action_cb((widget_t*)w, w->pathbuf, NULL);
}

/* fileselector <- widget */

static void shell_cb(widget_t *_w)
{
  filesel_widget_t *w = (filesel_widget_t*)(_w->parent);
  w->action_cb((widget_t*)w, w->pathbuf, NULL);
}

/*
 * Widget handling
 */

static long filesel_init (void)
{
	return 0;
}

static widget_t *filesel_create(widget_class_t *cp)
{
  filesel_widget_t *w = malloc (sizeof (filesel_widget_t));
  widget_t *vpane, *hpane, *option, *ok, *cancel;
  long a, b;

  if (!w)
    return NULL;
  memset (w, 0, sizeof (filesel_widget_t));
  w->w.class = wt_filesel_class;

  /* create the widget hierarchy */

  if(!(w->shell = wt_create(wt_shell_class, NULL)))
  {
    free(w);
    return NULL;
  }
  vpane     = wt_create(wt_pane_class,      w->shell);
  w->path   = wt_create(wt_getstring_class, vpane);
  w->name   = wt_create(wt_getstring_class, vpane);
  w->listbox= wt_create(wt_listbox_class,   vpane);
  w->info   = wt_create(wt_label_class,     vpane);
  hpane     = wt_create(wt_pane_class,      vpane);
  option    = wt_create(wt_button_class,    hpane);
  w->type   = wt_create(wt_getstring_class, hpane);
  ok        = wt_create(wt_button_class,    hpane);
  cancel    = wt_create(wt_button_class,    hpane);

  if(!(cancel && ok && option && hpane && vpane && w->type &&
       w->info && w->listbox && w->name && w->path))
  {
    wt_shell_class->delete(w->shell);
    free(w);
    return NULL;
  }
  /* add new hierarcy below fileselector */
  wt_add_before((widget_t*)w, NULL, w->shell);

  /* set options for widgets that will not change */

  a = AlignFill;
  wt_pane_class->setopt(vpane, WT_ALIGNMENT, &a);

  a = OrientHorz;
  wt_pane_class->setopt(hpane, WT_ORIENTATION, &a);

  wt_button_class->setopt(ok, WT_LABEL, "OK");
  wt_button_class->setopt(ok, WT_ACTION_CB, ok_cb);

  wt_button_class->setopt(cancel, WT_LABEL, "Cancel");
  wt_button_class->setopt(cancel, WT_ACTION_CB, cancel_cb);

  wt_button_class->setopt(option, WT_LABEL, "Options");
  wt_button_class->setopt(option, WT_ACTION_CB, option_cb);


  /* set options for widgets in filesel_widget_t struct */

  wt_shell_class->setopt(w->shell, WT_ACTION_CB, shell_cb);
  wt_shell_class->setopt(w->shell, WT_LABEL, "Fileselector");

  a = LabelModeWithBorder;
  wt_label_class->setopt(w->info, WT_MODE, &a);

  a = PATH_MAX;
  b = NAME_WIDTH;
  wt_setopt(w->path,
	WT_STRING_LENGTH, &a,
	WT_STRING_WIDTH, &b,
	WT_ACTION_CB, path_cb,
	WT_EOL);

  a = NAME_WIDTH;
  wt_setopt(w->name,
	WT_STRING_LENGTH, &a,
	WT_STRING_WIDTH, &a,
	WT_STRING_ADDRESS, "",
	WT_INKEY_CB, name_change_cb,
	WT_ACTION_CB, name_cb,
	WT_EOL);

  a = TYPE_WIDTH;
  wt_setopt(w->type,
	WT_STRING_LENGTH, &a,
	WT_STRING_WIDTH, &a,
	WT_STRING_ADDRESS, "*",
	WT_ACTION_CB, type_cb,
	WT_EOL);

  a = NAME_WIDTH;
  b = LIST_SHOWN;
  wt_setopt(w->listbox,
	WT_LIST_WIDTH, &a,
	WT_LIST_HEIGHT, &b,
	WT_INKEY_CB,  route_listbox_key,
	WT_CHANGE_CB, listbox_change_cb,
	WT_ACTION_CB, listbox_cb,
	WT_EOL);

  return (widget_t *)w;
}

static long filesel_delete(widget_t *_w)
{
  filesel_widget_t *w = (filesel_widget_t *)_w;

  wt_shell_class->delete(w->shell);
  free_entries(w);
  free(w);
  return 0;
}

static long filesel_close(widget_t *_w)
{
  filesel_widget_t *w = (filesel_widget_t *)_w;

  if(w->is_realized && w->is_open)
  {
    wt_shell_class->close(w->shell);
    w->is_open = 0;
  }
  return 0;
}

static long filesel_open(widget_t *_w)
{
  filesel_widget_t *w = (filesel_widget_t *)_w;

  if(w->is_realized && !w->is_open)
  {
    wt_shell_class->open(w->shell);
    w->is_open = 1;
  }
  return 0;
}

static long filesel_addchild(widget_t *parent, widget_t *w)
{
	return -1;
}

static long filesel_delchild(widget_t *parent, widget_t *w)
{
	return -1;
}

static long filesel_realize(widget_t *_w, WWIN *parent)
{
  filesel_widget_t *w = (filesel_widget_t *)_w;
  char *path;

  if(w->is_realized)
    return -1;

  wt_getstring_class->getopt(w->path, WT_STRING_ADDRESS, &path);

  if(!path)
    path = "";
  if(path[0] != '/')
    getcwd(w->pathbuf, PATH_MAX);

  if(!get_directory(w, path))
    return -1;

  if(wt_shell_class->realize(w->shell, parent) < 0)
    return -1;

  w->is_realized = 1;
  w->is_open = 1;
  return 0;
}

static long
filesel_query_geometry(widget_t *_w, long *xp, long *yp, long *wdp, long *htp)
{
  filesel_widget_t *w = (filesel_widget_t *)_w;

  wt_shell_class->query_geometry(w->shell, xp, yp, wdp, htp);
  return 0;
}

static long filesel_setopt (widget_t *_w, long key, void *val)
{
  filesel_widget_t *w = (filesel_widget_t *)_w;

  switch (key)
  {
    case WT_XPOS:
    case WT_YPOS:
    case WT_WIDTH:
    case WT_HEIGHT:
    case WT_LABEL:
      wt_shell_class->setopt(w->shell, key, val);
      break;

    case WT_LIST_HEIGHT:
      wt_listbox_class->setopt(w->listbox, key, val);
      break;

    case WT_FILESEL_PATH:
      if(w->is_realized)
        get_directory(w, (char*)val);
      else
        wt_getstring_class->setopt(w->path, WT_STRING_ADDRESS, val);
      break;

    case WT_FILESEL_MASK:
      wt_getstring_class->setopt(w->type, WT_STRING_ADDRESS, val);

      if(w->is_realized)
      {
        /* resort and show again */
        sort_entries(w, (char*)val);
        wt_listbox_class->setopt(w->listbox, WT_LIST_ADDRESS, w->dirinfo.start);
      }
      break;

    case WT_ACTION_CB:
      w->action_cb = val;
      break;

    default:
      return -1;
  }
  return 0;
}

static long filesel_getopt (widget_t *_w, long key, void *val)
{
  filesel_widget_t *w = (filesel_widget_t *)_w;

  switch (key)
  {
    case WT_XPOS:
    case WT_YPOS:
    case WT_WIDTH:
    case WT_HEIGHT:
    case WT_LABEL:
      wt_shell_class->getopt(w->shell, key, val);
      break;

    case WT_LIST_HEIGHT:
      wt_listbox_class->getopt(w->listbox, key, val);
      break;

    case WT_FILESEL_PATH:
      wt_listbox_class->getopt(w->path, WT_STRING_ADDRESS, val);
      break;

    case WT_FILESEL_MASK:
      wt_listbox_class->getopt(w->type, WT_STRING_ADDRESS, val);
      break;

    case WT_ACTION_CB:
      *(void **)val = w->action_cb;
      break;

    default:
      return -1;
  }
  return 0;
}

static long filesel_event (widget_t *_w, WEVENT *ev)
{
	return -1;
}

static long filesel_changes (widget_t *w, widget_t *w2, short changes)
{
	return 0;
}

static widget_class_t _wt_filesel_class = {
	"filesel", 0,
	filesel_init,
	filesel_create,
	filesel_delete,
	filesel_close,
	filesel_open,
	filesel_addchild,
	filesel_delchild,
	filesel_realize,
	filesel_query_geometry,
	filesel_setopt,
	filesel_getopt,
	filesel_event,
	filesel_changes,
	filesel_changes
};

widget_class_t *wt_filesel_class = &_wt_filesel_class;
