/*
 * this file is part of "The W Toolkit".
 *
 * (W) 1996, Eero Tamminen
 *
 * Getstring widget with scrolling and copy&paste for fixed width fonts.
 *
 * $Id: getstring.c,v 1.2 1996/04/04 20:11:10 yak Exp $
 *
 * Changes 25.3.1996 ++et,kay:
 * - Modified copy&paste also to do scrolling.
 * - Callback arguments changed.
 * Changes 28.3.1996 ++et:
 * - After realization only WT_STRING_ADDRESS can be setopt.
 *   (also fixed a related bug)
 * - Added a string change notify callback and a 'kill rest of line'
 *   shortcut.
 * - wt_timeout() can return 0. Use -1 for 'no timeout'.
 * - String shown at first so that the string end is visible.
 * - WT_INKEY_CB for reporting string changes.
 * - WT_STRING_MASK for filtering incoming characters.
 * - Changed the last argument on ACTION_CB from short to int.
 * - Widget same size with or without a border. Widget width
 *   configuarable, not just width in characters.
 * Changes 2.4.1996 ++et:
 * - String and mask are copied to internal, dynamically allocated space.
 * - Mask can now have ranges too.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <Wlib.h>
#include <Wt.h>
#include "toolkit.h"

/* Copy&Paste buffer. Will change into the W clipboard when it's ready */
#define BUFFER_SIZE	64
static char Buffer[BUFFER_SIZE+1];
static int BufferLenght;		/* whether there's something... */

/* special keys */
#define KEY_DEL		'\177'
#define KEY_LEFT	('B' & 0x1f)	/* emacs keys */
#define KEY_RIGHT	('F' & 0x1f)
#define KEY_HOME	('A' & 0x1f)
#define KEY_END		('E' & 0x1f)
#define KEY_DELREST	('K' & 0x1f)
#define KEY_CLEAR	'\33'		/* ESC clears the text */
/* destructive keys */
#define DESTRUCTIVE(k) (k == '\b' || \
	k == KEY_DEL || k == KEY_CLEAR || k == KEY_DELREST)

/* default font */
#define FONT_NAME	WT_DEFAULT_FONT
#define MAX_NAME_LEN	32
/* horizontal and vertical border offsets */
#define XBORDER		6
#define YBORDER		6
#define DEF_EDIT_WIDTH	32		/* default edit field width */

/* variables mean characters unless otherwise stated */
typedef struct
{
  widget_t w;
  short is_open;
  short is_realized;
  char fontname[MAX_NAME_LEN];
  WFONT *font;
  short mode;			/* GetstrNoBorder, GetstrWithBorder */
  short xborder;		/* horizontal border offset */
  char *text;			/* user buffer on which text is processed */
  char *mask;			/* allowed characters (NULL=all) */
  short bufsize;		/* current string buffer size */
  short maxlen;			/* the string buffer lenght */
  short textlen;		/* set / input text lenght */
  short showlen;		/* the string lenght on screen */
  short char_width;		/* fixed width character width */
  short cursor;			/* relative cursor position (in edit field) */
  short textpos;		/* string index to first char in edit box */
  short copystart;		/* painting starting postion (absolute) */
  short copyend;		/* painting ending position (absolute) */
  long  timer;			/* for copy&paste mouse checking */
  short cursor_on;
  /* call when a key is pressed / return is pressed */
  void (*key_cb) (widget_t *, char *text, int cursor);
  void (*ok_cb) (widget_t *, char *text, int cursor);
} getstring_widget_t;


/* show the character places on empty edit positions */
static void fill_sublines(getstring_widget_t *w)
{
  short idx = w->textlen - w->textpos;

  if (idx < w->showlen)
  {
    w_setmode(w->w.win, M_CLEAR);
    w_pbox(w->w.win,
      w->xborder + w->char_width * idx, YBORDER,
      w->char_width * (w->showlen - idx), w->font->height);
    w_setmode(w->w.win, M_INVERS);
    while(idx < w->showlen)
    {
      w_hline (w->w.win,
        w->xborder + w->char_width * idx + 1, YBORDER + w->font->height - 1,
	w->xborder + w->char_width * (idx+1) - 1);
      idx++;
    }
  }
}

/* draw the initial (everything postioned to zero)
 * text field, default graphics mode is M_INVERS
 */
static void
getstring_draw (getstring_widget_t * w)
{
  char bak;
  short idx;

  idx = w->textpos + w->showlen;
  idx = MIN(w->maxlen, idx);
  bak = w->text[idx];
  w->text[idx] = 0;
  w_printstring (w->w.win, w->xborder, YBORDER, &w->text[w->textpos]);
  w->text[idx] = bak;

  /* draw borders too if widget hasn't been already drawn */
  if (!w->is_realized && w->mode == GetstrModeWithBorder) {
    wt_box3d (w->w.win, 0, 0, w->w.w, w->w.h);
    wt_box3d_press (w->w.win, 0, 0, w->w.w, w->w.h, 0);
  }

  fill_sublines(w);
  w_setmode (w->w.win, M_INVERS);
}

static void
draw_to_endpos(getstring_widget_t * w)
{
  if((w->textpos = w->textlen - w->showlen) > 0)
    w->cursor = w->showlen;
  else
  {
    w->textpos = 0;
    w->cursor = w->textlen;
  }
  w->copystart = w->copyend = w->textpos + w->cursor;
  getstring_draw(w);
}

/* return total widget size */
static long
getstring_query_geometry (widget_t * _w, long *xp, long *yp, long *wdp, long *htp)
{
  getstring_widget_t *w = (getstring_widget_t *) _w;
  int def_width = w->char_width * w->showlen + 2 * w->xborder;

  if(!w->showlen)
    w->showlen = MIN (w->textlen, DEF_EDIT_WIDTH);
  else if(w->showlen > w->maxlen)
    w->showlen = w->maxlen;

  *xp = w->w.x;
  *yp = w->w.y;
  *htp = w->font->height + 2 * YBORDER;

  if(w->w.w > def_width)
  {
    def_width  = (w->w.w - 2*XBORDER) / w->char_width;
    w->showlen = MIN(w->maxlen, def_width);
    w->xborder = (w->w.w - w->char_width * w->showlen) >> 1;
    *wdp = w->char_width * w->showlen + 2 * w->xborder;
  }
  else
    *wdp = def_width;

  return 0;
}

/* scroll the text field contents to left */
static void
scroll_left(getstring_widget_t *w)
{
  if(w->textpos + w->showlen >= w->textlen)
    return;
  w->textpos++;
  w_bitblk(w->w.win, 
    w->xborder + w->char_width, YBORDER,
    w->char_width * (w->showlen - 1), w->font->height,
    w->xborder, YBORDER);
  w_printchar(w->w.win,
    w->xborder + w->char_width * (w->showlen-1), YBORDER,
    w->text[w->textpos + w->showlen - 1]);
  if(MAX(w->copystart, w->copyend) >= w->textpos + w->showlen &&
     MIN(w->copystart, w->copyend) <  w->textpos + w->showlen)
    w_pbox (w->w.win,
        w->xborder + w->char_width * (w->showlen-1), YBORDER,
        w->char_width, w->font->height);
}

/* scroll the text field contents to right */
static void
scroll_right(getstring_widget_t *w)
{
  if(w->textpos <= 0)
    return;
  w->textpos--;
  w_bitblk(w->w.win, w->xborder, YBORDER,
    w->char_width * (w->showlen - 1), w->font->height,
    w->xborder + w->char_width, YBORDER);
  w_printchar(w->w.win, w->xborder, YBORDER, w->text[w->textpos]);
  if(MIN(w->copystart, w->copyend) <= w->textpos &&
     MAX(w->copystart, w->copyend) >  w->textpos)
    w_pbox (w->w.win, w->xborder, YBORDER, w->char_width, w->font->height);
}

/* delete one character from the text string left of the cursor
 * and update the text field accordingly (scroll if needed).
 * check for whether there's something to delete is done elsewhere!
 */
static void
backspace_key(getstring_widget_t *w)
{
  short idx;

  /* move the rest of the characters */
  idx = w->textpos + w->cursor;
  while(idx <= w->textlen)
  {
    /* overlap, so better do manually, copy zero too... */
    w->text[idx-1] = w->text[idx];
    idx++;
  }
  w->textlen--;

  if(w->cursor <= 0)
  {
    w->textpos--;
    return;
  }
  /* need to move editfield contents? */
  if(w->textpos + w->cursor <= w->textlen && w->cursor < w->showlen)
  {
    idx = w->textlen - w->textpos - w->cursor + 1;
    if(idx > w->showlen - w->cursor)
      idx = w->showlen - w->cursor;
    w_bitblk(w->w.win, 
      w->xborder + w->char_width * w->cursor, YBORDER,
      w->char_width * idx, w->font->height,
      w->xborder + w->char_width * (w->cursor - 1), YBORDER);
  }
  /* need to replace last character with a new one? */
  if(w->textlen >= w->textpos + w->showlen)
    w_printchar(w->w.win,
      w->xborder + w->char_width * (w->showlen-1), YBORDER,
      w->text[w->textpos + w->showlen - 1]);
  else
  {
    /* or with a 'line'? */
    w_setmode(w->w.win, M_CLEAR);
    w_pbox(w->w.win,
      w->xborder + w->char_width * (w->textlen - w->textpos), YBORDER,
      w->char_width, w->font->height);
    w_setmode(w->w.win, M_INVERS);
    w_hline (w->w.win,
      w->xborder + w->char_width * (w->textlen - w->textpos) + 1,
      YBORDER + w->font->height - 1,
      w->xborder + w->char_width * (w->textlen - w->textpos + 1) - 1);
  }
  w->cursor--;
}

/* add one character to the text string right of the cursor
 * and update the text field accordingly (scroll if needed)
 */
static void
add_key(getstring_widget_t *w, short key)
{
  short idx;

  if(w->textlen >= w->maxlen)				/* enough space? */
    return;
 
  idx = w->textlen;
  while(idx >= w->textpos + w->cursor)
  {
    /* overlap, so better do manually, copy zero too... */
    w->text[idx+1] = w->text[idx];
    idx--;
  }
  w->text[w->textpos + w->cursor] = key;
  w->textlen++;

  if(w->cursor >= w->showlen)	/* at the end of editfield -> just scroll */
  {
    /* have to trick because editfield and string don't correlate before */
    scroll_left(w);
    return;
  }
  /* not at the text end? */
  if(w->textpos + w->cursor < w->textlen)
  {
    idx = w->textlen - w->textpos - w->cursor - 1;
    if(idx >= w->showlen - w->cursor)
      idx = w->showlen - w->cursor - 1;
    w_bitblk(w->w.win, 
      w->xborder + w->char_width * w->cursor, YBORDER,
      w->char_width * idx, w->font->height,
      w->xborder + w->char_width * (w->cursor + 1), YBORDER);
  }
  w_printchar(w->w.win, w->xborder + w->char_width * w->cursor, YBORDER, key);
  w->cursor++;
}
/* key masking.
 * Mask is composed of acceptable chars and char ranges.
 * "abcde", "0-9a-z", "-z" (upto z), "a-" (upwards from a)
 */
static inline short char_in(short key, char *mask)
{
  short cc, lc = 0;
  for(;;)
  {
    if(!(cc = *mask++))
      return 0;
    if(cc == '-')
    {
      if(lc <= key && (!*mask || key <= *mask++))
        return 1;
    }
    else if(key == (lc = cc))
      return 1;
  }
}

/* delegate key events */
static void process_key(getstring_widget_t *w, short key)
{
  short changed = 0;

  switch(key)
  {
    /* escape sequences */
    case '\b':
      if(w->cursor < 1 && w->textpos < 1)		/* nothing to do? */
        return;
      backspace_key(w);
      changed = 1;
      break;

    case KEY_DEL:
      if(w->textpos + w->cursor >= w->textlen)		/* nothing to do? */
        return;
      if(w->cursor < w->showlen)
      {
	w->cursor++;
	backspace_key(w);
      }
      else
      {
	/* no need to move editfield contents... */
	key = w->textpos + w->cursor;
	do
	{
	  w->text[key] = w->text[key+1];
	} while(++key < w->textlen);
	w->textlen--;
      }
      changed = 1;
      break;

    case KEY_DELREST:
      if(w->textpos + w->cursor >= w->textlen)		/* nothing to do? */
        return;
      /* delete to the end of the line */
      w->textlen = w->textpos + w->cursor;
      w->text[w->textlen] = 0;
      fill_sublines(w);
      changed = 1;
      break;

    case KEY_CLEAR:
      if(w->textlen < 1)				/* nothing to do? */
        return;
      memcpy(Buffer, w->text, w->textlen);		/* copy to buffer */
      w->text[0] = w->textpos = w->cursor = w->textlen = 0;
      fill_sublines(w);
      changed = 1;
      break;

    case KEY_LEFT:
      if(w->cursor > 0 || w->textpos > 0)
        if(w->cursor <= 0)
	  scroll_right(w);
	else
	  w->cursor--;
      break;

    case KEY_RIGHT:
      if(w->textpos + w->cursor < w->textlen)
        if(w->cursor >= w->showlen)
	  scroll_left(w);
	else
	  w->cursor++;
      break;

    case KEY_HOME:
      if(w->textpos == 0)
        w->cursor = 0;
      /*
       * this was w->cursor > 0. but w->cursor is the position in the visible
       * edit field, not in the string! kay.
       */
      if(w->textpos > 0)
      {
      	w->textpos = w->cursor = 0;
	getstring_draw(w);
      }
      break;

    case KEY_END:
      if(w->textpos + w->cursor < w->textlen)		/* something to do? */
        if(w->textpos + w->showlen < w->textlen)
          draw_to_endpos(w);
	else
	  w->cursor = w->textlen - w->textpos;
      break;

    case '\r':
    case '\n':
      if(w->ok_cb)
	w->ok_cb((widget_t *)w, w->text, w->textpos + w->cursor);
      break;

    default:
      /* printable? */
      if(key >= ' ' && (!w->mask || char_in(key, w->mask)))
	{
          add_key(w, key);
	  changed = 1;
	}
  }
  /* inform about text change? */
  if(w->key_cb && changed)
    w->key_cb((widget_t*)w, w->text, w->textpos + w->cursor);
}

/* show/remove cursor */
static void cursor_invert(getstring_widget_t *w)
{
  w->cursor_on ^= 1;
  w_vline (w->w.win, w->xborder + w->char_width * w->cursor, 3, w->w.h - 4);
}

/* paint text between x (relative) and copyend (absolute). set copyend to x */
static void paint_text(getstring_widget_t *w, short x)
{
  int max = MIN(w->showlen, w->textlen - w->textpos);
  int pos = w->copyend - w->textpos;

  /* need scrolling? */
  if(x < 0)
  {
    if(w->textpos > 0)
    {
      scroll_right(w);
      if(pos < max)
        pos++;
    }
    x = 0;
  }
  if(x > max)
  {
    if(w->textpos + w->showlen < w->textlen)
    {
      scroll_left(w);
      if(pos > 0)
        pos--;
    }
    x = max;
  }

  if (x != pos)
  {
    if (x > pos)
      w_pbox (w->w.win, w->xborder + w->char_width * pos,
         YBORDER, w->char_width * (x - pos), w->font->height);
    else
      w_pbox (w->w.win, w->xborder + w->char_width * x,
         YBORDER, w->char_width * (pos - x), w->font->height);
  }

  w->copyend = w->textpos + x;
}

/* process painting with mouse at intervals */
static void timer_callback (long _w)
{
  getstring_widget_t *w = (getstring_widget_t *) _w;
  short mx, my;

  if (w_querymousepos (w->w.win, &mx, &my) < 0)
    return;

  mx = (mx - w->xborder) / w->char_width;
  cursor_invert(w);
  paint_text(w, mx);
  cursor_invert(w);
  w->timer = wt_addtimeout (100, timer_callback, _w);
}

static void clear_selection(getstring_widget_t *w)
{
  if(w->copyend != w->copystart)
  {
    /* remove selection if any */
    if(w->copystart < w->textpos)
      paint_text(w, 0);
    else if(w->copystart > w->textpos + w->showlen)
      paint_text(w, w->showlen);
    else
      paint_text(w, w->copystart - w->textpos);
  }
}

static void remove_selection(getstring_widget_t *w)
{
  /* copyend is always visible (text scrolls with painting) */
  if(w->copyend < w->copystart)
  {
    memcpy(&w->text[w->copyend], &w->text[w->copystart],
      w->textlen - w->copystart + 1);
    w->textlen -= w->copystart - w->copyend;
    w->cursor = w->copyend - w->textpos;
  }
  else
  {
    memcpy(&w->text[w->copystart], &w->text[w->copyend],
      w->textlen - w->copyend + 1);
    w->textlen -= w->copyend - w->copystart;
    if(w->copystart < w->textpos)
    {
      w->textpos = w->copystart;
      w->cursor = 0;
    }
  }
  getstring_draw(w);
}

/* check cursor position etc. */
static void cursor_position(getstring_widget_t *w, short x)
{
  if(x + w->textpos > w->textlen)
    x = w->textlen - w->textpos;
  w->copystart = w->copyend = w->textpos + x;
  w->cursor = x;
}

/* process widget events */
static long getstring_event (widget_t * _w, WEVENT * ev)
{
  getstring_widget_t *w = (getstring_widget_t *) _w;
  short len, key, locate;

  switch (ev->type)
    {
    case EVENT_MPRESS:
      /* get keyboard focus */
      wt_getfocus (_w);

      /* cursor and selection off */
      cursor_invert(w);
      clear_selection(w);

      /* new cursor positions */
      cursor_position(w, (ev->x - w->xborder) / w->char_width);

      /* text painting (selection...) */
      if(ev->key == BUTTON_LEFT)
      {
        cursor_invert(w);
	timer_callback((long)_w);
      }
      else
        /* paste buffer to cursor position */
        if(ev->key == BUTTON_RIGHT && BufferLenght)
        {
	  len = -1;
	  while(++len < BufferLenght)
	    add_key(w, Buffer[len]);
	  cursor_position(w, w->cursor);
          cursor_invert(w);
	  if(w->key_cb)
	    /* inform about text change */
	    w->key_cb((widget_t*)w, w->text, w->textpos + w->cursor);
        }
      break;

    case EVENT_MRELEASE:
      if(w->timer >= 0)
      {
        wt_deltimeout(w->timer);
	w->timer = -1;
      }
      /* end painting and copy selected area to buffer */
      if(w->copyend != w->copystart && ev->key == BUTTON_LEFT)
      {
        /* copy painted text into internal (static) buffer */
	if(w->copyend < w->copystart)
	  len = w->copystart - w->copyend;
	else
	  len = w->copyend - w->copystart;
	len = MIN(BUFFER_SIZE, len);
        memcpy(Buffer, &w->text[MIN(w->copyend, w->copystart)], len);
	BufferLenght = len;
	Buffer[len] = 0;
      }
      break;

    case EVENT_KEY:
      key = ev->key;
      /* check for autolocator fed key events... */
      if(w->cursor_on)
      {
        cursor_invert(w);
	locate = 0;
      }
      else
        locate = 1;

      /* remove selection if needed */
      if(w->copyend != w->copystart)
      {
        if(key == KEY_LEFT || key == KEY_RIGHT ||
	   key == KEY_HOME || key == KEY_END)
	  clear_selection(w);
        else
	{
	  remove_selection(w);

	  /* just the selection? */
	  if(DESTRUCTIVE(key))
	  {
	    if(w->key_cb)
	      /* inform about text change */
	      w->key_cb((widget_t*)w, w->text, w->textpos + w->cursor);
	    cursor_position(w, w->cursor);
            if(!locate)
	      cursor_invert(w);
	    return 0;
	  }
	}
      }
      /* insert a key (changes cursor position) */
      process_key(w, key);
      cursor_position(w, w->cursor);
      if(!locate)
        cursor_invert(w);
      break;
    }
  return 0;
}

/* ------------------------- */

static long
getstring_init (void)
{
  return 0;
}

static int
getstring_loadfont (getstring_widget_t *w, char *fname)
{
  WFONT *fp = w_loadfont (fname);
  if (!fp)
    return -1;
  w->font = fp;
  strncpy (w->fontname, fname, sizeof (w->fontname));
  w->fontname[sizeof (w->fontname) - 1] = '\0';
  w->char_width = w_strlen (w->font, " ");
  return 0;
}

static widget_t *
getstring_create (widget_class_t * cp)
{
  getstring_widget_t *wp = malloc (sizeof (getstring_widget_t));
  if (!wp)
    return NULL;
  memset (wp, 0, sizeof (getstring_widget_t));
  wp->w.class = wt_getstring_class;
  if (getstring_loadfont (wp, FONT_NAME))
    {
      free (wp);
      return NULL;
    }
  wp->timer = -1;
  wp->mode = GetstrModeWithBorder;
  wp->xborder = XBORDER;

  return (widget_t *) wp;
}

static long
getstring_delete (widget_t * _w)
{
  getstring_widget_t *w = (getstring_widget_t *) _w;

  wt_deltimeout (w->timer);
  if (w->is_realized)
    w_delete (w->w.win);
  w_unloadfont (w->font);
  if(w->text)
    free(w->text);
  if(w->mask)
    free(w->mask);
  free (w);
  return 0;
}

static long
getstring_close (widget_t * _w)
{
  getstring_widget_t *w = (getstring_widget_t *) _w;

  if (w->is_realized && w->is_open)
    {
      w_close (w->w.win);
      w->is_open = 0;
    }
  return 0;
}

static long
getstring_open (widget_t * _w)
{
  getstring_widget_t *w = (getstring_widget_t *) _w;

  if (w->is_realized && !w->is_open)
    {
      w_open (w->w.win, w->w.x, w->w.y);
      w->is_open = 1;
    }
  return 0;
}

static long
getstring_addchild (widget_t * parent, widget_t * w)
{
  return -1;
}

static long
getstring_delchild (widget_t * parent, widget_t * w)
{
  return -1;
}

static long
getstring_realize (widget_t * _w, WWIN * parent)
{
  getstring_widget_t *w = (getstring_widget_t *) _w;
  long x, y, wd, ht;

  if (w->is_realized || !(w->text && w->maxlen))
    return -1;

  getstring_query_geometry (_w, &x, &y, &wd, &ht);
  w->w.x = x;
  w->w.y = y;
  w->w.w = wd;
  w->w.h = ht;

  w->w.win = wt_create_window (parent, wd, ht,
	W_NOBORDER | W_MOVE | EV_KEYS | EV_MOUSE, _w);
  if (!w->w.win)
    return -1;
  w->w.win->user_val = (long) w;

  w_setfont (w->w.win, w->font);
  /* cursor to last character & redraw the widget */
  draw_to_endpos (w);
  w->is_realized = 1;
  w->is_open = 1;
  w_open (w->w.win, w->w.x, w->w.y);
  return 0;
}

static long
getstring_setopt (widget_t * _w, long key, void *val)
{
  getstring_widget_t *w = (getstring_widget_t *) _w;
  short len, mask = 0;
  char *tmp;

  switch (key)
    {
    case WT_XPOS:
      w->w.x = *(long *) val;
      if (w->is_realized)
      {
        w_move (w->w.win, w->w.x, w->w.y);
        mask |= WT_CHANGED_POS;
      }
      break;

    case WT_YPOS:
      w->w.y = *(long *) val;
      if (w->is_realized)
      {
	w_move (w->w.win, w->w.x, w->w.y);
	mask |= WT_CHANGED_POS;
      }
      break;

    /* mask notification needed only when realized... */
    case WT_WIDTH:
      if(!w->is_realized)
        w->w.w = *(long *) val;
      break;

    case WT_HEIGHT:
      if(!w->is_realized)
        w->w.h = *(long *) val;
      break;

    case WT_FONT:
      if(!w->is_realized)
      {
        if (getstring_loadfont (w, (char *) val))
	  return -1;
        w_setfont (w->w.win, w->font);
      }
      break;

    case WT_STRING_ADDRESS:
      len = strlen ((char*)val);
      if(!w->text || len >= w->bufsize)
      {
	if(!(tmp = malloc(len+1)))
	  return -1;
	w->bufsize = len+1;
	if(w->text)
	  free(w->text);
	w->text = tmp;
      }
      strcpy(w->text, (char*)val);
      w->textlen = len;
      if(!w->maxlen)
	w->maxlen = len;
      if (w->is_realized)
      {
        short cursor_on = w->cursor_on;
	if (cursor_on)
	  cursor_invert (w);
	draw_to_endpos (w);
	if (cursor_on)
	  cursor_invert (w);
	if(w->timer >= 0)
	{
	  wt_deltimeout(w->timer);
	  w->timer = -1;
	}
      }
      break;

    case WT_STRING_LENGTH:
      w->maxlen = *(long *) val;
      if(w->maxlen >= w->bufsize)
      {
	if((tmp = malloc(w->maxlen+1)))
	{
	  w->bufsize = w->maxlen+1;
	  if(w->text)
	  {
	    strcpy(tmp, w->text);
	    free(w->text);
	  }
	  w->text = tmp;
	}
        else
	  w->maxlen = (w->bufsize > 0 ? w->bufsize : 0);
      }
      if(w->showlen > w->maxlen)
        w->showlen = w->maxlen;
      break;

    case WT_STRING_MASK:
      len = strlen((char*)val);
      if(!w->mask || strlen(w->mask) < len)
      {
	if(!(tmp = malloc(len+1)))
	  return -1;
	if(w->mask)
	  free(w->mask);
	w->mask = tmp;
      }
      strcpy(w->mask, (char*)val);
      break;

    case WT_STRING_WIDTH:
      if(w->is_realized)
        return -1;
      w->showlen = *(long *) val;
      break;

    case WT_INKEY_CB:
      w->key_cb = val;
      break;

    case WT_ACTION_CB:
      w->ok_cb = val;
      break;

    case WT_MODE:
      if (w->is_realized)
        return -1;
      w->mode = *(long *)val;
      break;

    default:
      return -1;
    }
  if (mask)
    wt_change_notify (_w, mask);

  return 0;
}

static long
getstring_getopt (widget_t * _w, long key, void *val)
{
  getstring_widget_t *w = (getstring_widget_t *) _w;

  switch (key)
    {
    case WT_XPOS:
      *(long *) val = w->w.x;
      break;

    case WT_YPOS:
      *(long *) val = w->w.y;
      break;

    case WT_WIDTH:
      *(long *) val = w->w.w;
      break;

    case WT_HEIGHT:
      *(long *) val = w->w.h;
      break;

    case WT_FONT:
      *(char **) val = w->fontname;
      break;

    case WT_STRING_ADDRESS:
      *(char **) val = w->text;
      break;

    case WT_STRING_MASK:
      *(char **) val = w->mask;
      break;

    case WT_STRING_LENGTH:
      *(long *) val = w->maxlen;
      break;

    case WT_STRING_WIDTH:
      *(long *) val = w->showlen;
      break;

    case WT_INKEY_CB:
      *(void **)val = w->key_cb;
      break;

    case WT_ACTION_CB:
      *(void **)val = w->ok_cb;
      break;

    case WT_MODE:
      *(long *)val = w->mode;
      break;

    default:
      return -1;
    }
  return 0;
}

static long
getstring_changes (widget_t * w, widget_t * w2, short changes)
{
  return 0;
}

static long
getstring_focus (widget_t *_w, int enter)
{
  getstring_widget_t *w = (getstring_widget_t *)_w;

  cursor_invert (w);
  if (!enter && w->timer >= 0) {
    wt_deltimeout (w->timer);
    w->timer = -1;
  }
  return 0;
}

static widget_class_t _wt_getstring_class =
{
  "getstring", 0,
  getstring_init,
  getstring_create,
  getstring_delete,
  getstring_close,
  getstring_open,
  getstring_addchild,
  getstring_delchild,
  getstring_realize,
  getstring_query_geometry,
  getstring_setopt,
  getstring_getopt,
  getstring_event,
  getstring_changes,
  getstring_changes,
  getstring_focus
};

widget_class_t *wt_getstring_class = &_wt_getstring_class;
