/*
 * this file is part of "The W Toolkit".
 *
 * (W) 1996, Eero Tamminen from Kay's button widget.
 *
 * label/infobox widget.
 *
 * $Id: label.c,v 1.2 1996/04/06 00:08:06 yak Exp $
 *
 * - Align options: top, bottom, left, right, center.
 * - Same widget size and text offset regardless of the border.
 * - Before realization returns the minimum size.
 * - Widget can have both user drawn stuff (eg. icon) and text.
 *   They better be differently aligned and widget wide enough...
 * - Widget text is redrawn in most economical way. Shouldn't
 *   matter whether font's a fixed or proportional one.
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <Wlib.h>
#include <Wt.h>
#include "toolkit.h"

/* text offsets from window border */
#define XBORDER		WT_DEFAULT_XBORDER
#define YBORDER		WT_DEFAULT_YBORDER

typedef struct {
	widget_t w;
	char *label;
	char fontname[32];
	WFONT *font;
	short is_open;
	short is_realized;
	short mode;
	short halign;
	short valign;
	short textlen;
	short oldwidth;
	void (*draw_fn) (widget_t *w, long x, long y, long wd, long ht);
} label_widget_t;


static void
redraw_contents (label_widget_t *w)
{
	short xoffset, yoffset, border, width, height;
	WWIN *win = w->w.win;

	if (w->draw_fn)
		(*w->draw_fn) ((widget_t *)w, XBORDER, YBORDER,
			w->w.w - 2*XBORDER, w->w.h - 2*YBORDER);

	if (!(w->label && (width = w_strlen (w->font, w->label))))
		return;

	height = w->font->height;

	/* text placement */
	switch (w->valign) {
	case AlignTop:
		yoffset = YBORDER;
		break;
	case AlignBottom:
		yoffset = w->w.h - height - YBORDER;
		break;
	default:
		yoffset = (w->w.h - height) >> 1;
	}
	switch (w->halign) {
	case AlignLeft:
		xoffset = XBORDER;
		break;
	case AlignRight:
		xoffset = w->w.w - width - XBORDER;
		break;
	default:
		xoffset = (w->w.w - width) >> 1;
	}

	border = MAX (w->oldwidth - width, 0);
	/* need to clear old text? */
	if (border > 0) {
		w_setmode (win, M_CLEAR);
		switch (w->halign) {
		case AlignLeft:
			w_pbox (win, xoffset + width, yoffset,
				border, height);
			break;
		case AlignRight:
			w_pbox (win, w->w.w - w->oldwidth, yoffset,
				border, height);
			break;
		default:
			border = (border >> 1) + 1;
			w_pbox (win, xoffset-border, yoffset, border, height);
			w_pbox (win, xoffset+width, yoffset, border, height);
		}
		w_setmode (win, M_DRAW);
	}

	w_printstring (win, xoffset, yoffset, w->label);
	w->oldwidth = width;
}

static void
label_draw (label_widget_t *w)
{
	WWIN *win = w->w.win;

	w_setmode (win, M_CLEAR);
	w_pbox (win, 0, 0, w->w.w, w->w.h);
	w_setmode(win, M_DRAW);

	if (w->mode == LabelModeWithBorder)
		w_box(win, 0, 0, w->w.w, w->w.h);

	redraw_contents (w);
}

static long
label_query_geometry (widget_t *_w, long *xp, long *yp, long *wdp, long *htp)
{
	label_widget_t *w = (label_widget_t *)_w;
	int wd, ht, textwidth;

	*xp = w->w.x;
	*yp = w->w.y;
	*wdp = w->w.w;
	*htp = w->w.h;

	textwidth = w_strlen (w->font, w->label);
	wd = textwidth + 2*XBORDER;
	ht = w->font->height + 2*YBORDER;

	if(*wdp < wd)
		*wdp = wd;
	if(*htp < ht)
		*htp = ht;

	return 0;
}

static int
label_loadfont (label_widget_t *w, char *fname)
{
	WFONT *fp = w_loadfont (fname);
	if (!fp)
		return -1;
	w->font = fp;
	strncpy (w->fontname, fname, sizeof (w->fontname));
	w->fontname[sizeof (w->fontname) - 1] = '\0';
	return 0;
}

	
static long
label_init (void)
{
	return 0;
}

static widget_t *
label_create (widget_class_t *cp)
{
	label_widget_t *wp = malloc (sizeof (label_widget_t));
	if (!wp)
		return NULL;
	memset (wp, 0, sizeof (label_widget_t));
	wp->w.class = wt_label_class;

	/* defaults */
	wp->mode = LabelModeNoBorder;
	wp->w.w = 2*XBORDER;
	wp->w.h = 2*YBORDER;

	if (label_loadfont (wp, WT_DEFAULT_FONT)) {
		free (wp);
		return NULL;
	}
	return (widget_t *)wp;
}

static long
label_delete (widget_t *_w)
{
	label_widget_t *w = (label_widget_t *)_w;

	if (w->is_realized)
		w_delete (w->w.win);
	w_unloadfont (w->font);
	if (w->label)
		free (w->label);
	free (w);
	return 0;
}

static long
label_close (widget_t *_w)
{
	label_widget_t *w = (label_widget_t *)_w;

	if (w->is_realized && w->is_open) {
		w_close (w->w.win);
		w->is_open = 0;
	}
	return 0;
}

static long
label_open (widget_t *_w)
{
	label_widget_t *w = (label_widget_t *)_w;

	if (w->is_realized && !w->is_open) {
		w_open (w->w.win, w->w.x, w->w.y);
		w->is_open = 1;
	}
	return 0;
}

static long
label_addchild (widget_t *parent, widget_t *w)
{
	return -1;
}

static long
label_delchild (widget_t *parent, widget_t *w)
{
	return -1;
}

static long
label_realize (widget_t *_w, WWIN *parent)
{
	label_widget_t *w = (label_widget_t *)_w;
	long x, y, wd, ht;

	if (w->is_realized)
		return -1;

	label_query_geometry (_w, &x, &y, &wd, &ht);
	w->w.x = x;
	w->w.y = y;
	w->w.w = wd;
	w->w.h = ht;

	w->w.win = wt_create_window (parent, wd, ht, W_NOBORDER|W_MOVE,	_w);
	if (!w->w.win)
		return -1;
	w->w.win->user_val = (long)w;

	w_setfont (w->w.win, w->font);
	label_draw (w);
	w->is_realized = 1;
	w->is_open = 1;
	w_open (w->w.win, w->w.x, w->w.y);
	return 0;
}

static long
label_setopt (widget_t *_w, long key, void *val)
{
	label_widget_t *w = (label_widget_t *)_w;
	short len, mask = 0;
	int needredraw = 0;
	char *tmp;

	switch (key) {
	case WT_XPOS:
		w->w.x = *(long *)val;
		mask |= WT_CHANGED_POS;
		if (w->is_realized) {
			w_move (w->w.win, w->w.x, w->w.y);
		}
		break;

	case WT_YPOS:
		w->w.y = *(long *)val;
		mask |= WT_CHANGED_POS;
		if (w->is_realized) {
			w_move (w->w.win, w->w.x, w->w.y);
		}
		break;

	case WT_WIDTH:
		w->w.w = *(long *)val;
		mask |= WT_CHANGED_SIZE;
		break;

	case WT_HEIGHT:
		w->w.h = *(long *)val;
		mask |= WT_CHANGED_SIZE;
		break;

	case WT_LABEL:
		if (val) {
			len = strlen ((char*)val);
			if (len >= w->textlen)
			{
				if (!(tmp = malloc (len+1)))
					return -1;
				w->textlen = len+1;
				if (w->label)
					free (w->label);
				w->label = tmp;
			}
			strcpy (w->label, (char*)val);
			if (w->is_realized) {
				redraw_contents (w);
			}
		} else {
			if (w->label)
				free (w->label);
			w->label = NULL;
		}
		return 0;

	case WT_FONT:
		if (label_loadfont (w, (char *)val)) {
			return -1;
		}
		w_setfont (w->w.win, w->font);
		needredraw = 1;
		break;

	case WT_DRAW_FN:
		w->draw_fn = val;
		needredraw = 1;
		break;

	case WT_MODE:
		w->mode = *(long *)val;
		needredraw = 1;
		break;

	case WT_ALIGNMENT:
		len = *(long *)val;
		if (len == AlignCenter) {
			w->valign = AlignCenter;
			w->halign = AlignCenter;
		} else {
			if (len == AlignTop || len == AlignBottom)
				w->valign = len;
			else
				w->halign = len;
		}
		needredraw = 1;
		break;

	default:
		return -1;
	}
	if (mask && w->is_realized)
		wt_change_notify (_w, mask);
	if (w->is_realized && needredraw) {
		label_draw (w);
	}
	return 0;
}

static long
label_getopt (widget_t *_w, long key, void *val)
{
	label_widget_t *w = (label_widget_t *)_w;

	switch (key) {
	case WT_XPOS:
		*(long *)val = w->w.x;
		break;

	case WT_YPOS:
		*(long *)val = w->w.y;
		break;

	case WT_WIDTH:
		*(long *)val = w->w.w;
		if (!w->is_realized && w->label) {
			*(long *)val += w_strlen (w->font, w->label);
		}
		break;

	case WT_HEIGHT:
		*(long *)val = w->w.h;
		if (!w->is_realized) {
			*(long *)val += w->font->height;
		}
		break;

	case WT_LABEL:
		*(char **)val = w->label;
		break;

	case WT_FONT:
		*(char **)val = w->fontname;
		break;

	case WT_DRAW_FN:
		*(void **)val = w->draw_fn;
		break;

	case WT_MODE:
		*(long *)val = w->mode;
		break;

	case WT_ALIGNMENT:
		*(long *)val = w->halign | w->valign;
		break;

	default:
		return -1;
	}
	return 0;
}

static long
label_event (widget_t *_w, WEVENT *ev)
{
	return -1;
}

static long
label_changes (widget_t *w, widget_t *w2, short changes)
{
	return 0;
}

static widget_class_t _wt_label_class = {
	"label", 0,
	label_init,
	label_create,
	label_delete,
	label_close,
	label_open,
	label_addchild,
	label_delchild,
	label_realize,
	label_query_geometry,
	label_setopt,
	label_getopt,
	label_event,
	label_changes,
	label_changes
};

widget_class_t *wt_label_class  = &_wt_label_class;
