/*
 * this file is part of "The W Toolkit".
 *
 * (W) 1996, Eero Tamminen.
 *
 * listbox widget.
 *
 * $Id: listbox.c,v 1.2 1996/04/04 20:11:16 yak Exp $
 *
 * Shows a (vertically) scrollable list of strings of which one
 * at the time can be selected.
 *
 * For the time being, the given WT_LIST_ADDRESS list is accessed from
 * the application space.
 *
 * Widget WT_LIST_WIDTH or WT_LIST_HEIGHT attributes (in characters, width
 * will work correctly only for fixed with fonts of course) cannot be
 * changed after widget is realized.
 *
 * The total widget width is the largest of WT_WIDTH, WT_LIST_WIDTH
 * or the lenght of the longest list string (+ borders & scrollbar).
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <Wlib.h>
#include <Wt.h>
#include "toolkit.h"

/* keyboard shortcuts (emacs) */
#define KEY_UP		('P' & 0x1f)	/* _p_revious line */
#define KEY_DOWN	('N' & 0x1f)	/* _n_ext line */
#define KEY_PGUP	('B' & 0x1f)	/* _b_acwards */
#define KEY_PGDN	('F' & 0x1f)	/* _f_orwards */
#define KEY_HOME	('A' & 0x1f)
#define KEY_END		('E' & 0x1f)

/* for 3D looks... */
#define OFFSET		2			/* border for text invert box */
#define XBORDER		(WT_DEFAULT_XBORDER+4)	/* horizontal listbox border */
#define YBORDER		(WT_DEFAULT_YBORDER+1)	/* vertical -"- */

/* default font */
#define FONT_NAME	WT_DEFAULT_FONT
#define MAX_NAME_LEN	32

typedef struct
{
  widget_t w;
  short is_open;
  short queried;
  short is_realized;
  char fontname[MAX_NAME_LEN];
  widget_t *scrollbar;
  WFONT *font;
  short xborder;		/* horizontal border offset */
  short height;			/* list box height in chars */
  short chars;			/* list box width in chars */
  short width;			/* list box width in pixels */
  short items;			/* how many strings there is */
  char **list;			/* NULL terminated string list */
  short top;			/* first visible string */
  short current;		/* currently selected string */
  /* called when item is clicked / shortcuts are used / char unknown */
  void (*select_cb)(widget_t *, char *text, int idx);
  void (*change_cb)(widget_t *, char *text, int idx);
  void (*key_cb)(widget_t *, WEVENT *ev);
} listbox_widget_t;


static void
cursor_invert(listbox_widget_t *w)
{
  int idx = w->current - w->top;
  if(idx >= 0 && idx < w->height)
  {
    w_pbox(w->w.win,
      w->xborder-OFFSET, YBORDER + idx * w->font->height,
      w->width - (w->xborder-OFFSET)*2, w->font->height);
  }
}

/* draw the texts (other stuff is done straight in realize)
 * if not flag, clear the whole are before drawing strings
 */
static long
listbox_draw (listbox_widget_t *w, int flag)
{
  WWIN *win = w->w.win;
  short y, width, idx;

  if(w->top > w->items - w->height)
    w->top = w->items - w->height;

  y = YBORDER;
  w_setmode(win, M_CLEAR);
  if(!flag)
    w_pbox(win, w->xborder, YBORDER, w->width - w->xborder*2, w->w.h - YBORDER*2);
  for(idx = w->top; idx < w->top + w->height; idx++)
  {
    if(flag)
    {
      /* less flickering than clearing the whole string width */
      width = w->width - w->xborder*2 - w_strlen(w->font, w->list[idx]);
      if(width > 0)
	w_pbox(win, w->width - w->xborder - width, y, width, w->font->height);
    }
    w_printstring(win, w->xborder, y, w->list[idx]);
    y += w->font->height;
  }
  w_setmode(win, M_INVERS);
  return 0;
}

static long
listbox_query_geometry (widget_t * _w, long *xp, long *yp, long *wdp, long *htp)
{
  listbox_widget_t *w = (listbox_widget_t *) _w;
  static long extrawidth;
  int size;

  /* if size is too small, fix it so that everything fits */
  *xp = w->w.x;
  *yp = w->w.y;
  *wdp = w->w.w;
  *htp = w->w.h;
  if(!w->list)
    return -1;

  if(w->queried)
  {
    /* only width can be be enlargened */
    size = w->width + extrawidth;
    if(*wdp > size)
    {
      w->xborder += (*wdp - size) >> 1;
      w->width = *wdp - extrawidth;
    }
    else
      *wdp = size;
    return 0;
  }
  w->queried = 1;

  w->width = w->items = 0;
  (*wt_scrollbar_class->getopt) (w->scrollbar, WT_WIDTH, &extrawidth);
  extrawidth += 2;	/* distance between text box and scrollbar */

  /* window width in fixed point chars */
  if(w->chars)
  {
    /* works reliably only for fixed with chars!!! */
    w->width = w->font->widths[32] * w->chars;
    while(w->list[w->items])
      w->items++;
  }
  else
  {
    /* get longest string width (in pixels) */
    while(w->list[w->items])
    {
      if((size = w_strlen(w->font, w->list[w->items])) > w->width)
	w->width = size;
      w->items++;
    }
  }
  w->width += w->xborder*2;
  size = w->width + extrawidth;
  /* select widest */
  if(*wdp > size)
  {
    w->xborder += (*wdp - size) >> 1;
    w->width = *wdp - extrawidth;
  }
  else
    *wdp = size;

  if(w->height < 1)
    w->height = w->items;		/* show all */

  size = w->height * w->font->height + YBORDER*2;
  if(*htp < size)
    *htp = size;
  else
    w->height = (*htp - YBORDER*2) / w->font->height;

  if(w->height > w->items)
    w->height = w->items;

  w->w.w = *wdp;
  w->w.h = *htp;
  return 0;
}

static void
scroll_up(listbox_widget_t *w)
{
  WWIN *win = w->w.win;
  int width, x, y;

  if(w->height < 2)
    return;

  y = (w->height-1) * w->font->height;
  w_bitblk(win,
    w->xborder, YBORDER + w->font->height,
    w->width - w->xborder*2, y,
    w->xborder, YBORDER);

  y += YBORDER;
  x = w->top + w->height;
  w_printstring(win, w->xborder, y, w->list[x]);
  width = w_strlen(w->font, w->list[x-1]);
  x = w_strlen(w->font, w->list[x]);
  width -= x;
  if(width > 0)
  {
    w_setmode(win, M_CLEAR);
    w_pbox(win, w->xborder + x, y, width, w->font->height);
    w_setmode(win, M_INVERS);
  }
  w->top++;
}

static void
scroll_down(listbox_widget_t *w)
{
  WWIN *win = w->w.win;
  int width, x;

  if(w->height < 2)
    return;

  w_bitblk(w->w.win,
    w->xborder, YBORDER,
    w->width - w->xborder*2, (w->height-1) * w->font->height,
    w->xborder, YBORDER + w->font->height);

  w->top--;
  w_printstring(win, w->xborder, YBORDER, w->list[w->top]);
  x = w_strlen(w->font, w->list[w->top]);
  width = w_strlen(w->font, w->list[w->top+1]) - x;
  if(width > 0)
  {
    w_setmode(win, M_CLEAR);
    w_pbox(win, w->xborder + x, YBORDER, width, w->font->height);
    w_setmode(win, M_INVERS);
  }
}

static void
listbox_redraw(listbox_widget_t *w, short newtop)
{
  if(newtop-1 == w->top)
  {
    scroll_up(w);
    return;
  }
  if(newtop+1 == w->top)
  {
    scroll_down(w);
    return;
  }
  w->top = newtop;
  /* redraw completely */  
  listbox_draw(w, 1);
}

static void
scroll_callback (widget_t *_w, int pos, int pressed)
{
  listbox_widget_t *w = (listbox_widget_t *) _w->parent;
  short newtop;

  if((newtop = pos) == w->top)
    return;

  cursor_invert(w);
  listbox_redraw(w, newtop);
  cursor_invert(w);
}

static long
listbox_init (void)
{
  return 0;
}

static int
listbox_loadfont (listbox_widget_t *w, char *fname)
{
  WFONT *fp = w_loadfont (fname);
  if (!fp)
    return -1;
  w->font = fp;
  strncpy (w->fontname, fname, sizeof (w->fontname));
  w->fontname[sizeof (w->fontname) - 1] = '\0';
  return 0;
}

static widget_t *
listbox_create (widget_class_t *cp)
{
  int i;
  listbox_widget_t *wp = malloc (sizeof (listbox_widget_t));
  if (!wp)
    return NULL;

  memset (wp, 0, sizeof (listbox_widget_t));
  wp->w.class = wt_listbox_class;
  if (listbox_loadfont (wp, FONT_NAME))
    {
      free (wp);
      return NULL;
    }

  wp->scrollbar = wt_create (wt_scrollbar_class, NULL);
  if (!wp->scrollbar)
  {
    w_unloadfont (wp->font);
    free (wp);
    return NULL;
  }
  wt_add_before ((widget_t*)wp, NULL, wp->scrollbar);

  (*wt_scrollbar_class->setopt) (wp->scrollbar, WT_ACTION_CB, scroll_callback);
  (*wt_scrollbar_class->getopt) (wp->scrollbar, WT_HEIGHT, &i);
  wp->w.h = i;

  wp->current = -1;
  wp->xborder = XBORDER;
  return (widget_t *) wp;
}

static long
listbox_delete (widget_t * _w)
{
  listbox_widget_t *w = (listbox_widget_t *) _w;

  if (w->is_realized)
    w_delete (w->w.win);
  wt_delete (w->scrollbar);
  w_unloadfont (w->font);
  free (w);
  return 0;
}

static long
listbox_close (widget_t * _w)
{
  listbox_widget_t *w = (listbox_widget_t *) _w;

  if (w->is_realized && w->is_open)
    {
      w_close (w->w.win);
      w->is_open = 0;
    }
  return 0;
}

static long
listbox_open (widget_t * _w)
{
  listbox_widget_t *w = (listbox_widget_t *) _w;

  if (w->is_realized && !w->is_open)
    {
      w_open (w->w.win, w->w.x, w->w.y);
      w->is_open = 1;
    }
  return 0;
}

/* needed for the wt_create(scrollbar)... */
static long
listbox_addchild (widget_t *_w, widget_t *child)
{
  return -1;
}

static long
listbox_delchild (widget_t *_parent, widget_t *w)
{
  return -1;
}

static void
set_scrollbar(listbox_widget_t *w)
{
  long i;

  if(w->items <= w->height)
  {
    i = w->height;
    (*wt_scrollbar_class->setopt) (w->scrollbar, WT_TOTAL_SIZE, &i);
    (*wt_scrollbar_class->setopt) (w->scrollbar, WT_SIZE, &i);

    i = 0;
    (*wt_scrollbar_class->setopt) (w->scrollbar, WT_POSITION, &i);
    (*wt_scrollbar_class->setopt) (w->scrollbar, WT_PAGE_INC, &i);
    (*wt_scrollbar_class->setopt) (w->scrollbar, WT_LINE_INC, &i);
  }
  else
  {
    i = w->items;
    (*wt_scrollbar_class->setopt) (w->scrollbar, WT_TOTAL_SIZE, &i);

    i = w->top;
    (*wt_scrollbar_class->setopt) (w->scrollbar, WT_POSITION, &i);

    i = w->height;
    (*wt_scrollbar_class->setopt) (w->scrollbar, WT_SIZE, &i);
    (*wt_scrollbar_class->setopt) (w->scrollbar, WT_PAGE_INC, &i);

    i = 1;
    (*wt_scrollbar_class->setopt) (w->scrollbar, WT_LINE_INC, &i);
  }
}

static long
listbox_realize (widget_t * _w, WWIN * parent)
{
  listbox_widget_t *w = (listbox_widget_t *) _w;
  long y;

  if (w->is_realized || !(w->list && w->font))
    return -1;

  /* modifies variables straigh */
  listbox_query_geometry (_w, &y, &y, &y, &y);
  y = w->w.x + w->width + 2;
  /* setup scrollbar eometry and attributes */
  (*wt_scrollbar_class->setopt) (w->scrollbar, WT_XPOS, &y);
  (*wt_scrollbar_class->setopt) (w->scrollbar, WT_YPOS, &w->w.y);
  (*wt_scrollbar_class->setopt) (w->scrollbar, WT_HEIGHT, &w->w.h);

  w->w.win = wt_create_window (parent, w->width, w->w.h,
	W_MOVE | W_NOBORDER | EV_MOUSE | EV_KEYS, _w);
  if (!w->w.win)
    return -1;

  w_setfont (w->w.win, w->font);
  wt_box3d (w->w.win, 0, 0, w->width, w->w.h);
  wt_box3d_press (w->w.win, 0, 0, w->width, w->w.h, 0);
  /* draw texts and setopt scrollbar */
  listbox_draw (w, 0);
  set_scrollbar (w);
  (*wt_scrollbar_class->realize) (w->scrollbar, parent);

  w_open (w->w.win, w->w.x, w->w.y);
  w_setmode (w->w.win, M_INVERS);
  w->w.win->user_val = (long) w;
  w->is_realized = 1;
  w->is_open = 1;
  return 0;
}

static void
set_cursorpos(listbox_widget_t *w, long cursor)
{
  int start;

  if(w->current >= 0)
    cursor_invert(w);

  if(cursor < 0)
  {
    w->current = -1;
    return;
  }
  if(cursor >= w->items)
    cursor = w->items - 1;

  start = w->top;
  w->current = cursor;
  if(cursor < start || cursor >= start + w->height)
  {
    /* has to change list 'window' position */
    start = cursor;
    if(start > w->items - w->height)
      start = w->items - w->height;
    /* redraw */
    listbox_redraw(w, start);
    set_scrollbar(w);
  } 
  /* set cursor */
  cursor_invert(w);
}

static void
show_new_list(listbox_widget_t *w)
{
  cursor_invert(w);

  w->items = 0;
  while(w->list[w->items])
    w->items++;

  w->height = (w->w.h - YBORDER*2) / w->font->height;
  if(w->height > w->items)
    w->height = w->items;
  w->current = -1;

  /* app should take care that strings still fit into window */
  listbox_draw (w, 0);
  set_scrollbar(w);
}

static long
listbox_setopt (widget_t * _w, long key, void *val)
{
  listbox_widget_t *w = (listbox_widget_t *) _w;
  short mask = 0;

  switch (key)
    {
    case WT_XPOS:
      w->w.x = *(long *) val;
      mask |= WT_CHANGED_POS;
      if (w->is_realized)
	w_move (w->w.win, w->w.x, w->w.y);
      break;

    case WT_YPOS:
      w->w.y = *(long *) val;
      mask |= WT_CHANGED_POS;
      if (w->is_realized)
	w_move (w->w.win, w->w.x, w->w.y);
      break;

    /* mask notification needed only when realized... */
    case WT_WIDTH:
      if(!w->is_realized)
        w->w.w = *(long *) val;
      break;

    case WT_HEIGHT:
      if(!w->is_realized)
        w->w.h = *(long *) val;
      break;

    case WT_FONT:
      if(!w->is_realized)
      {
	if (listbox_loadfont (w, (char *) val))
	  return -1;
	w_setfont (w->w.win, w->font);
      }
      break;

    case WT_LIST_WIDTH:
      /* works only for fixed with fonts!!! */
      if(!w->is_realized)
        w->chars = *(long *) val;
      break;

    case WT_LIST_HEIGHT:
      if(!w->is_realized)
        w->height = *(long *) val;
      break;

    case WT_LIST_ADDRESS:
      w->list = (char **) val;
      if(w->is_realized)
        show_new_list(w);
      break;

    case WT_CURSOR:
      if(w->is_realized)
        set_cursorpos(w, *(long *)val);
      break;

    case WT_INKEY_CB:
      w->key_cb = val;
      break;

    case WT_ACTION_CB:
      w->select_cb = val;
      break;

    case WT_CHANGE_CB:
      w->change_cb = val;
      break;

    default:
      return -1;
    }

  if (mask && w->is_realized)
    wt_change_notify (_w, mask);

  return 0;
}

static long
listbox_getopt (widget_t * _w, long key, void *val)
{
  listbox_widget_t *w = (listbox_widget_t *) _w;

  switch (key)
    {
    case WT_XPOS:
      *(long *) val = w->w.x;
      break;

    case WT_YPOS:
      *(long *) val = w->w.y;
      break;

    case WT_WIDTH:
      *(long *) val = w->w.w;
      break;

    case WT_HEIGHT:
      *(long *) val = w->w.h;
      break;

    case WT_FONT:
      *(char **) val = w->fontname;
      break;

    case WT_LIST_HEIGHT:
      *(long *) val = w->height;
      break;

    case WT_LIST_ADDRESS:
      *(char ***) val = w->list;
      break;

    case WT_CURSOR:
      *(long *)val = w->current;
      break;

    case WT_INKEY_CB:
      *(void **)val = w->key_cb;
      break;

    case WT_ACTION_CB:
      *(void **)val = w->select_cb;
      break;

    case WT_CHANGE_CB:
      *(void **)val = w->change_cb;
      break;

    default:
      return -1;
    }
  return 0;
}

static long
listbox_event (widget_t * _w, WEVENT * ev)
{
  listbox_widget_t *w = (listbox_widget_t *) _w;
  short new, old, tmp;

  switch (ev->type)
  {
    case EVENT_MPRESS:
      /* get keyboard focus */
      wt_getfocus (_w);

      /* current selection / new selection on listbox */
      old = w->current - w->top;
      new = (ev->y - YBORDER) / w->font->height;

      if(new != old && new >= 0 && new < w->height &&
         ev->x > 0 && ev->x < w->width)
      {
        /* deselect old */
	cursor_invert(w);
	/* select new */
	w->current = new + w->top;
	cursor_invert(w);

	/* If there were a double click mouse event, I could add it to do
	 * selection return.  Single click would then signify just cursor
	 * position change.  I could of course do double click check with
	 * timeouts, but that would need changing when W server itself gets
	 * double clicks...
	 */
	if(w->select_cb)
	  w->select_cb((widget_t *)w, w->list[w->current], w->current);
      }
      return 0;


    case EVENT_KEY:
      new = -1;
      old = w->current;
      switch(ev->key)
      {
        case KEY_UP:
	  if(w->current < 0)
	    w->current = w->top + w->height - 1;
	  else if(w->current > 0)
	    w->current--;
	  break;

	case KEY_DOWN:
	  if(w->current < 0)
	    w->current = w->top;
	  else if(w->current+1 < w->items)
	    w->current++;
	  break;

	case KEY_PGUP:
	  new = w->top - w->height;
	  if(new < 0)
	    new = 0;
	  if(w->current < 0)
	    w->current = new + w->height - 1;
	  else
	  {
	    w->current -= w->height;
	    if(w->current < 0)
	      w->current = 0;
	  }
	  break;

	case KEY_PGDN:
	  new = w->top + w->height;
	  if(new > w->items - w->height)
	    new = w->items - w->height;
	  if(w->current < 0)
	    w->current = new;
	  else
	  {
	    w->current += w->height;
	    if(w->current >= w->items)
	      w->current = w->items-1;
	  }
	  break;
	  
	case KEY_HOME:
	  w->current = 0;
          break;

	case KEY_END:
	  w->current = w->items - 1;
	  break;

	case '\n':
	case '\r':
	  if(w->select_cb && w->current >= 0)
	    w->select_cb((widget_t *)w, w->list[w->current], w->current);
	  return 0;

	default:
	  if(w->key_cb)
	    /* let app wonder about the key... */
	    w->key_cb((widget_t *)w, ev);
	  return 0;
      }
      /* no need to draw anything? */
      if(new < 0 && old == w->current)
        return 0;

      /* hide old cursor */
      tmp = w->current;
      w->current = old;
      cursor_invert(w);
      w->current = tmp;

      if((new >= 0 && new != w->top) ||
         tmp < w->top || tmp >= w->top + w->height)
      {
        long pos;

        if(new < 0)
	{
	  /* calculate new list 'window' position */
	  if(tmp == w->top + w->height)
	    new = w->top + 1;
	  else
	    new = tmp;
	  if(new > w->items - w->height)
	    new = w->items - w->height;
	}
	/* redraw */
	listbox_redraw(w, new);

	pos = new;
	wt_setopt(w->scrollbar, WT_POSITION, &pos, WT_EOL);

      }
      /* show new cursor and notify about cursor position change */
      cursor_invert(w);
      if(w->change_cb)
        w->change_cb((widget_t *)w, w->list[w->current], w->current);
      return 0;
  }
  return 0;
}

static long
listbox_changes (widget_t * w, widget_t * w2, short changes)
{
  return 0;
}

static long
listbox_focus (widget_t *_w, int enter)
{
  return 0;
}

static widget_class_t _wt_listbox_class =
{
  "listbox", 0,
  listbox_init,
  listbox_create,
  listbox_delete,
  listbox_close,
  listbox_open,
  listbox_addchild,
  listbox_delchild,
  listbox_realize,
  listbox_query_geometry,
  listbox_setopt,
  listbox_getopt,
  listbox_event,
  listbox_changes,
  listbox_changes,
  listbox_focus
};

widget_class_t *wt_listbox_class = &_wt_listbox_class;
